#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

# Logic for searching for nodes and networks within the library path and
# generating a mouse menu.
#
# TODO: possibly make the library a global for all canvases


import os
import shutil
import subprocess
from functools import partial

# gpi
from gpi import QtCore, QtGui, QtWidgets
from .config import Config
from .defaultTypes import GPITYPE_PASS
from .defines import isWidget, isGPIType, isExternalNode
from .catalog import Catalog, CatalogObj
from .defaultTypes import GPIDefaultType
from .defines import isGPIModFile, isGPITypeFile, isGPINetworkFile, GPI_PYMOD_PRE_EXT
from .loader import loadMod, PKGroot, appendSysPath
from .node import Node
from .sysspecs import Specs

from .logger import manager

# start logger for this module
log = manager.getLogger(__name__)

NOPATH_MESSAGE = "<em>No library selected...</em>"

class SearchMenu(QtWidgets.QMenu):
    '''A menu class that leaves keyboard focus with its parent.'''
    def __init__(self, menuPos, parent=None):
        super().__init__(parent=parent)
        self._menuPos = menuPos
        self._parent = parent

    def show(self):
        self.popup(self._menuPos)
        self._parent.grabKeyboard()

    def close(self):
        super().close()
        self._parent.releaseKeyboard()

class NodeCatalogItem(CatalogObj):
    '''A single entry to a Node database. For information such as library,
    path, type, etc...
    '''

    def __init__(self, fullpath):

        dn = os.path.dirname
        bn = os.path.basename

        self.fullpath = fullpath

        # determine if there is an editable code file for direct node edit access
        self.editable_path = None
        epath, ext = os.path.splitext(fullpath)
        epath += '.py'
        if os.path.isfile(epath):
            self.editable_path = epath

        # 'SpiralCoords_GPI.py' and path
        fil = bn(fullpath)
        path = dn(fullpath)

        # get module name and lib name
        name, ext = os.path.splitext(fil)  # 'SpiralCoords_GPI' '.py'

        # verify the '_GPI' before the file extension
        self.isNodeFile = False
        if name.endswith(GPI_PYMOD_PRE_EXT):
            self.isNodeFile = True

        name = name.split(GPI_PYMOD_PRE_EXT)[0]  # get node's display name 'SpiralCoords'
        second = bn(path)  # 'GPI' if the GPI dir is used
        third = bn(dn(path))  # 'core' if GPI-dir is NOT used
        lib_path = dn(path)

        # skip the GPI directory name by removing it from path
        if second == 'GPI':
            # 'spiral'
            second = bn(dn(path))
            # 'core'
            third = bn(dn(dn(path)))
            lib_path = dn(dn(path))

        # save node info
        self.name = name
        self.second = second
        self.third = third
        self._lib_path = lib_path
        self.path = path
        self.ext = [ext]

        self.pkg_root = None
        if self.path:
            self.pkg_root = PKGroot(self.path)

        # This is the only element that cannot be duplicated across nodes
        self._id = self.third+'.'+self.second+'.'+self.name
        self.thrd_sec = self.third+'.'+self.second

        # add to sys path in class node module has pkg.module refs
        if self.pkg_root:
            appendSysPath(self.pkg_root)

        # try loading the node module
        self.mod = None
        self.widgetNames = []
        #self.load()

    def key(self):
        return self._id

    def lib(self):
        return self.third

    def lib_path(self):
        return self._lib_path

    def sameLib(self, item):
        return item.lib_path() == self.lib_path()

    def libWarn(self, item):
        msg = 'Multiple library paths with the name: \''+str(self.lib())+'\'\n' \
                + 'Paths:\n\t* ' + str(self.lib_path()) +'\n\t  '+ str(item.lib_path()) \
                + '\nNode:\n\t  ' + str(self.key())+' '+str(item.ext)+'\n' \
                + 'Skipping duplicate library...'
        log.warn(msg)

    def find(self, key):
        return (key in self.types)

    def reload(self):
        log.info("reload: "+str(self.fullpath))
        self.load()

    def load(self):

        # clear previous instance
        self.mod = None

        # enforce filename law
        if not self.isNodeFile:
            log.error('Not a valid GPI node filename, skipping. ('+str(self.fullpath)+')')
            return

        # verify that there is a node description
        mod = loadMod(self.fullpath)
        if mod:
            if hasattr(mod, 'ExternalNode'):
                if isExternalNode(getattr(mod, 'ExternalNode')):
                    self.mod = mod
                else:
                    log.error('ExternalNode object found, but it is not a GPI node def. ('+str(self.fullpath)+')')
                    return
            else:
                log.error('No ExternalNode definition found, skipping. ('+str(self.fullpath)+')')
                return
        else:
            log.error('Module not loadable, skipping. ('+str(self.fullpath)+')')
            return

        if self.mod:
            # check if it has widget definitions
            self.widgetNames = []  # re-init list in case of reload
            for name in dir(self.mod):
                if isWidget(getattr(self.mod, name)):
                    self.widgetNames.append(name)

            if len(self.widgetNames):
                log.info(str(self.fullpath)+' contains widget definitions:  '+str(self.widgetNames))

    def valid(self):
        return self.isLoaded() # and self.isNodeFile

    def validExt(self):
        return self.isNodeFile

    def isLoaded(self):
        if self.mod:
            return True
        else:
            return False

    def hasWidgets(self):
        return len(self.widgetNames) > 0

    def getWidget(self, name):
        if name in self.widgetNames:
            return getattr(self.mod, name)

    def appendExt(self, ext):
        self.ext.append(ext)
        self.ext = list(set(self.ext))

    def merge(self, m):

        if not self.sameLib(m):
            self.libWarn(m)
            return

        # this handles collisions between nodes found with the same id
        self.ext += m.ext
        self.ext = list(set(self.ext))

    def description(self):
        # return the node description class
        return self.mod.ExternalNode

    def __str__(self):
        return '< '+self._id+', '+str(self.path)+', '+str(self.ext) +' >'

class NetworkCatalogItem(CatalogObj):
    '''A single database entry to a GPI Network database.
    '''

    def __init__(self, fullpath):

        # TODO: more rigorous file checking (perhaps a magic number).

        dn = os.path.dirname
        bn = os.path.basename

        self.fullpath = fullpath

        # 'SpiralCoords_GPI.net' and path
        fil = bn(fullpath)
        path = dn(fullpath)

        # get module name and lib name
        name, ext = os.path.splitext(fil)  # 'SpiralCoords_GPI' '.net'

        # verify the '_GPI' before the file extension
        # determine whether to install net in the library menu
        self.isNetLibFile = False
        if name.endswith(GPI_PYMOD_PRE_EXT):
            self.isNetLibFile = True

        name = name.split(GPI_PYMOD_PRE_EXT)[0]  # get node's display name 'SpiralCoords'
        second = bn(path)  # 'GPI' if the GPI dir is used
        third = bn(dn(path))  # 'core' if GPI-dir is NOT used

        # skip the GPI directory name by removing it from path
        if second == 'GPI':
            # 'spiral'
            second = bn(dn(path))
            # 'core'
            third = bn(dn(dn(path)))

        # save net info
        self.name = name
        self.second = second
        self.third = third
        self.path = path
        self.ext = ext

        # This is the only element that cannot be duplicated across nodes
        self._id = self.third+'.'+self.second+'.'+self.name
        self.thrd_sec = self.third+'.'+self.second

    def valid(self):
        return isGPINetworkFile(self.fullpath) and self.isNetLibFile

    def key(self):
        return self._id

    def appendExt(self, ext):
        self.ext.append(ext)
        self.ext = list(set(self.ext))

    def merge(self, m):
        # this handles collisions between nodes found with the same id
        self.ext += m.ext
        self.ext = list(set(self.ext))

    def __str__(self):
        return self._id+', '+str(self.path)+', '+str(self.ext)

class GPITYPECatalogItem(CatalogObj):
    '''Object specific to GPITYPE plugins.
    1) input the full path: /path-to-plugin-dir/mytype_GPITYPE.py or .pyc
    2) this will try to verify the path
    3) try to load the python module
    4) failures will be flagged
    '''

    def __init__(self, fullpath):

        if type(fullpath) is not str:
            log.error('GPITYPECatalogItem requires a string argument.')

        self.fullpath = fullpath
        self.name, ext = os.path.splitext(fullpath)
        self.path, self.name = os.path.split(self.name)
        self.ext = [ext]

        # unique name
        self._id = self.path+'/'+self.name

        # try to have python load the module
        self.mod = loadMod(fullpath)

        # Extract the GPITYPES from the module for checking and keep a list of names.
        self.types = []
        for cn in dir(self.mod):
            cls = getattr(self.mod, cn)
            if isGPIType(cls):
                log.info('\t'+fullpath+' has GPITYPE: '+cn)
                self.types.append(cn)

    def valid(self):
        return self.isLoaded()

    def isLoaded(self):
        if self.mod:
            return True
        else:
            return False

    def key(self):
        return self._id

    def appendExt(self, ext):
        self.ext.append(ext)
        self.ext = list(set(self.ext))

    def merge(self, m):
        # this handles collisions between nodes found with the same id
        self.ext += m.ext
        self.ext = list(set(self.ext))

    def __str__(self):
        return self._id+', '+str(self.path)+', '+str(self.ext)

    def hasType(self, key):
        # GPITYPE libraries can hold multiple types. Usually, the type is what
        # is searched for.
        return (key in self.types)

    def type(self, key):
        # return the type object associated with the key name.
        if self.hasType(key):
            return getattr(self.mod, key)()

class Library(object):
    '''Contains all the Node, Network, and GPIType path searching, mouse menu
    generation and indexing for the node library.  The contents of the library
    are loaded at startup (each time) and when the user adds Nodes via drag'n
    drop or menu contexts.
    '''

    def __init__(self, parent):
        self._parent = parent  # must be a Qt parent for signalling
        self._known_GPI_nodes = Catalog()  # list of all modules within each lib
        self._known_GPI_networks = Catalog()  # all networks in each lib
        self._known_GPI_types = Catalog() # all GPI types found in init search
        self.extTypes = dict()
        self._listwdg = None  # for searching node list

        self.generateNewNodeListWindow()

        self._lib_menus = {}  # third level menu (holds second lev list)
        self._lib_second = {}  # second level menu (holds node list)
        self._lib_menu = []  # third level menu list

        self.scanGPIModulesIn_LibraryPath(recursion_depth=3)
        self.generateLibMenus()
        self.generateNewNodeList()

    def showNewNodeListWindow(self):
        self._list_win.show()

    def _get_new_node_name(self):
        fname = self._new_node_name_field.text()
        if fname == "":
            fname = self._new_node_name_field.placeholderText()
        if not fname.endswith(GPI_PYMOD_PRE_EXT + '.py'):
            fname += GPI_PYMOD_PRE_EXT + '.py'
        return fname

    def _createNewNode(self):
        # copy node template to this library, and open it up
        fullpath = self._new_node_path

        new_node_created = False
        if os.path.exists(fullpath):
            log.warn("Didn't create new node at path: " + fullpath +
                     " (file already exists)")
        else:
            try:
                shutil.copyfile(Config.GPI_NEW_NODE_TEMPLATE_FILE,
                                fullpath)
            except OSError as e:
                print(e)
                log.warn("Didn't create new node at path: " + fullpath)
            else:
                log.dialog("New node created at path: " + fullpath)
                new_node_created = True
                self.rescan()

        self._list_win.hide()

        if new_node_created:
            # instantiate our new node on the canvas
            canvas = self._parent
            pos = QtCore.QPoint(0, 0)
            node = self.findNode_byPath(fullpath)
            sig = {'sig': 'load', 'subsig': node, 'pos': pos}
            canvas.addNodeRun(sig)

            # now open the file for editing (stolen from node.py)
            if Specs.inOSX():
                # OSX users set their launchctl associated file prefs
                command = "open \"" + fullpath + "\""
                subprocess.Popen(command, shell=True)
            # Linux users set their editor choice
            # TODO: this should be moved to config
            elif Specs.inLinux():
                editor = 'gedit'
                if os.environ.has_key("EDITOR"):
                    editor = os.environ["EDITOR"]
                command = editor + " \"" + fullpath + "\""
                subprocess.Popen(command, shell=True)
            else:
                log.warn("Quick-Edit unavailable for this OS, aborting...")

    def _setQTLabelElided(self, label, text):
        fm = QtGui.QFontMetrics(label.font())
        width = label.width()
        elided_text = fm.elidedText(text, QtCore.Qt.ElideMiddle, width)
        label.setText(elided_text)

    def _newNodeNameEdited(self):
        new_name = self._get_new_node_name()
        current_path = self._new_node_path
        if current_path != '':
            path, old_name = os.path.split(current_path)
            fullpath = os.path.join(path, new_name)
            self._new_node_path = fullpath
            self._setQTLabelElided(self._new_node_path_field, fullpath)

    # This slot is called whenever a list item is clicked. This is used to
    # update the path and set the enabled/disabled state of the create node
    # button.
    def _listItemClicked(self, item):
        idx, label = self._new_node_list_index
        if idx == 0:
            self._create_button.setDisabled(True)
            self._new_node_path_field.setText(NOPATH_MESSAGE)
            self._new_node_path = ''
        elif idx == 1:
            if item.text() == '..':
                self._create_button.setDisabled(True)
                self._new_node_path_field.setText(NOPATH_MESSAGE)
                self._new_node_path = ''
            else:
                for k in self._known_GPI_nodes.keys():
                    node = self._known_GPI_nodes.get(k)
                    if node.thrd_sec == '.'.join((label, item.text())):
                        fullpath = os.path.join(node.path, self._get_new_node_name())
                        self._new_node_path = fullpath
                        self._setQTLabelElided(self._new_node_path_field, fullpath)
                        self._create_button.setEnabled(True)
                        break
        elif idx == 2:
            self._create_button.setEnabled(True)

    # This slot is called whenever a list item is double-clicked. This is used
    # for navigation of the library lists when creating a new node.
    def _listItemDoubleClicked(self, item):
        new_node_created = False

        idx, label = self._new_node_list_index
        if idx == 0:
            self.generateNewNodeList(item.text())
        elif item.text() == '..':
            new_index = label.split('.')
            if idx == 1:
                self.generateNewNodeList()
            else:
                self.generateNewNodeList('.'.join(new_index[:idx-1]))
        elif idx < 2:
            new_index = '.'.join((label, item.text()))
            self.generateNewNodeList(new_index)

    def scanForNewNodes(self):
        log.dialog("Scanning for newly created modules and libraries...")
        self.scanGPIModulesIn_LibraryPath(recursion_depth=3)
        self.regenerateLibMenus()
        log.dialog("Finished rescanning.")

    def rescan(self):

        # remove all libs for a fresh rescan
        self._known_GPI_nodes = Catalog()  # list of all modules within each lib
        self._known_GPI_networks = Catalog()  # all networks in each lib
        self._known_GPI_types = Catalog() # all GPI types found in init search
        self.extTypes = dict()

        log.dialog("Rescanning for newly created modules and libraries...")
        self.scanGPIModulesIn_LibraryPath(recursion_depth=3)
        self.regenerateLibMenus()
        self.generateNewNodeList()
        log.dialog("Finished rescanning.")

    def getUserLibsWithPaths(self):
        return None

    def getType(self, key):
        # GPITYPE
        # if requested type was returned, then include a True, else False.

        if key == GPITYPE_PASS:
            return (GPIDefaultType(), True)

        typObj = self._known_GPI_types.intrafind('hasType', key)
        if typObj is None:
            log.info('Requested port-type: \''+str(key)+'\' not found.  Using \'' + str(GPITYPE_PASS) + '\' instead.')
            return (GPIDefaultType(), False)
        else:
            return (typObj.type(key), True)

    def findNode_byName(self, name):
        # expose the catalog's find function
        return self._known_GPI_nodes.find('name', name)

    def findNode_byPath(self, path):
        return self._known_GPI_nodes.find('fullpath', path)

    def findNode_byLibrary(self, name, second, third):
        key = third+'.'+second+'.'+name
        if key in list(self._known_GPI_nodes.keys()):
            return self._known_GPI_nodes.get(key)

    def findNode_byKey(self, key):
        if key in list(self._known_GPI_nodes.keys()):
            return self._known_GPI_nodes.get(key)

    def findNode_byClosestMatch(self, name, wdg_port_names):
        # find all nodes with the same name, then try to match as many widget
        # and port names from the list.
        log.debug('Find node by closest match: '+str(name))

        byname = self._known_GPI_nodes.list('name', name)

        if byname is None:
            log.warn('No node found with name \''+str(name)+'\'')
            return None

        if len(byname) == 0:
            log.warn(str(name) + ' node not found.')
            return None

        if len(byname) == 1:
            log.info('Found one match for '+str(name))
            return byname[0]

        any_cnt = []
        exact_cnt = []
        for item in byname:

            # need an instance to check the node's widget and port names
            cur_names = Node(None, nodeCatItem=item).getWidgetAndPortNames()

            # try checking for any match type
            cnt = 0
            for wpn in wdg_port_names:
                match = False
                for cn in cur_names:
                    # if one is in the other or the other is in the one
                    if cn.lower().count(wpn.lower()) or wpn.lower().count(cn.lower()):
                        match = True
                if match:
                    cnt += 1
            any_cnt.append(cnt*100.0/len(wdg_port_names))

            # try checking for exact match type
            cnt = 0
            for wpn in wdg_port_names:
                if wpn in cur_names:
                    cnt += 1
            exact_cnt.append(cnt*100.0/len(wdg_port_names))

        msg = 'Find Node: '+str(name)+'\n'
        msg += 'Candidates:\n'
        for item, cpc, epc in zip(byname, any_cnt, exact_cnt):
            msg += '\t'+str(item.fullpath) + ' any: '+str(cpc)+'%, exact: '+str(epc)+'%\n'

        if len(set(exact_cnt)) > 1:
            msg += 'Exact Name Match:\n'

            chosen_item = byname[0]
            chosen_score = exact_cnt[0]
            for item, epc in zip(byname, exact_cnt):
                if chosen_score < epc:
                    chosen_score = epc
                    chosen_item = item

            msg +='\tChosen: ' + str(item.fullpath) + ', MP: ' + str(chosen_score) + '%'
            log.warn(msg)
            return chosen_item

        if len(set(any_cnt)) > 1:
            msg += 'Any Name Match:\n'

            chosen_item = byname[0]
            chosen_score = any_cnt[0]
            for item, epc in zip(byname, any_cnt):
                if chosen_score < epc:
                    chosen_score = epc
                    chosen_item = item

            msg += '\tChosen: ' + str(item.fullpath) + ', MP: ' +str(chosen_score)+'%'
            log.warn(msg)
            return chosen_item

        log.warn(msg)
        log.warn('No node distinction found, choosing first match: '+str(byname[0].fullpath))
        return byname[0]

    def addNode(self, item, check=True):
        if check:
            item.load()
            if not item.valid():
                return -1

        existing = self.findNode_byKey(item.key())
        if not existing:
            self._known_GPI_nodes.append(item)
            return 1
        elif not existing.sameLib(item):
            existing.libWarn(item)
            return -1
        return 0 # already in database

    def listWdg(self):
        return self._listwdg

    def libMenus(self):
        return self._lib_menus

    def libMenu(self):
        return self._lib_menu

    def searchMenu(self, txt, parent, mousemenu):
        # this menu needs to be rebuilt everytime a character changes

        # don't bother if the user just cleared the search
        if len(txt) == 0:
            return

        pos = self._parent.mapToGlobal(self._parent._event_pos + QtCore.QPoint(mousemenu.sizeHint().width(), 0))

        # close any existing search menu and assign the new one
        self.removeSearchPopup()

        searchMenu = SearchMenu(pos, parent=parent)
        self.generateNodeSearchActions(str(txt), searchMenu, mousemenu)

        self._listwdg = searchMenu
        searchMenu.show()

    def removeSearchPopup(self):
        if self._listwdg is not None:
            self._listwdg.close()
            self._listwdg = None

    def addNodeAndCloseMouseMenu(self, s, searchmenu, mousemenu):
        '''Close all menus and related objects.
        NOTE: THIS ORDER MUST BE PRESERVED!!!
        '''
        self.removeSearchPopup()
        mousemenu.close()
        self._parent.activateWindow()
        self._parent.update()
        self._parent.addNodeRun(s)

    def scanGPIModulesIn_LibraryPath(self, recursion_depth=1):
        new_sys_paths = []
        types_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'types')
        for spath in Config.GPI_LIBRARY_PATH + [types_path]:
            path = os.path.realpath(spath)  # remove excess '/'
            if os.path.isdir(path):
                self.scanGPIModules(path, recursion_depth)

        log.info("GPI mods/nets/types found:")
        log.info(str(self._known_GPI_nodes))
        log.info(str(self._known_GPI_networks))
        log.info(str(self._known_GPI_types))
        log.info('number of nodes: '+str(len(self._known_GPI_nodes)))
        log.info('number of networks: '+str(len(self._known_GPI_networks)))
        log.info('number of types: '+str(len(self._known_GPI_types)))

    # TODO: move this and others like it to a common help-object that can errorcheck.
    def openGPIRCHelp(self):
        if not QtGui.QDesktopServices.openUrl(QtCore.QUrl('http://docs.gpilab.com')):
            QtWidgets.QMessageBox.information(self, 'Documentation',"Documentation can be found at\nhttp://docs.gpilab.com", QtWidgets.QMessageBox.Close)

    # http://docs.gpilab.com/Configuration/#configuration-library-directories
    def openLIBDIRSHelp(self):
        if not QtGui.QDesktopServices.openUrl(QtCore.QUrl('http://docs.gpilab.com/Configuration/#configuration-library-directories')):
            QtWidgets.QMessageBox.information(self, 'Documentation',"Documentation can be found at\nhttp://docs.gpilab.com", QtWidgets.QMessageBox.Close)

    def regenerateLibMenus(self):
        self._lib_menus = {}  # third level menu (holds second lev list)
        self._lib_second = {}  # second level menu (holds node list)
        self._lib_menu = []  # third level menu list
        self.generateLibMenus()
        self.generateNewNodeList()

    def generateLibMenus(self):
        # default menu if no libraries are found
        numnodes = len(list(self._known_GPI_nodes.keys()))
        if numnodes == 0:
            self._lib_menus['No Nodes Found'] = QtWidgets.QMenu('No Nodes Found')
            buf = 'Check your ~/.gpirc for the correct LIB_DIRS.'
            act = QtWidgets.QAction(buf, self._parent, triggered = self.openLIBDIRSHelp)
            self._lib_menus['No Nodes Found'].addAction(act)

            for m in sorted(list(self._lib_menus.keys()), key=lambda x: x.lower()):
                mm = self._lib_menus[m]
                mm.setTearOffEnabled(False)
                self._lib_menu.append(mm)

            return

        # NODE MENU
        # setup libs using node id. ex: core.mathematics.sum
        # the ids of
        for k in sorted(self._known_GPI_nodes.keys(), key=lambda x: x.lower()):
            node = self._known_GPI_nodes.get(k)
            if node.third not in self._lib_menus:
                #self._lib_menus[node.third] = QtWidgets.QMenu(node.third.capitalize())
                self._lib_menus[node.third] = QtWidgets.QMenu(node.third)
                self._lib_menus[node.third].setTearOffEnabled(True)

            if node.thrd_sec not in self._lib_second:
                self._lib_second[node.thrd_sec] = QtWidgets.QMenu(node.second)
                self._lib_second[node.thrd_sec].setTearOffEnabled(True)
                ma = self._lib_menus[node.third].addMenu(self._lib_second[node.thrd_sec])

            sm = self._lib_second[node.thrd_sec]

            # TODO: try setting up hotkeys/shortcuts for specific nodes
            a = QtWidgets.QAction(node.name, self._parent, statusTip="Click to instantiate the \'"+str(node.name)+"\' node.")
            s = {'subsig': node}
            a.triggered.connect(partial(self._parent.addNodeRun, sig=s))
            sm.addAction(a)

        # NETWORK MENU
        for sm in list(self._lib_second.values()):
            sm.addSeparator()

        for k in sorted(list(self._known_GPI_networks.keys()), key=lambda x: x.lower()):
            net = self._known_GPI_networks.get(k)
            if net.third not in self._lib_menus:
                #self._lib_menus[net.third] = QtWidgets.QMenu(net.third.capitalize())
                self._lib_menus[net.third] = QtWidgets.QMenu(net.third)
                self._lib_menus[net.third].setTearOffEnabled(True)

            if net.thrd_sec not in self._lib_second:
                self._lib_second[net.thrd_sec] = QtWidgets.QMenu(net.second)
                self._lib_second[net.thrd_sec].setTearOffEnabled(True)
                self._lib_menus[net.third].addMenu(self._lib_second[node.thrd_sec])

            sm = self._lib_second[net.thrd_sec]
            a = QtWidgets.QAction(net.name + ' (net)', self._parent, statusTip="Click to instantiate the \'"+str(net.name)+"\' network.")
            s = {'sig': 'load', 'subsig': 'net', 'path': net.fullpath}
            a.triggered.connect(partial(self._parent.addNodeRun, sig=s))
            sm.addAction(a)

        for m in sorted(list(self._lib_menus.keys()), key=lambda x: x.lower()):
            mm = self._lib_menus[m]
            mm.setTearOffEnabled(True)
            self._lib_menu.append(mm)

    # each time this is called it goes through all the nodes to populate the
    # list
    def generateNewNodeList(self, top_lib=None):
        list_items = set()
        new_node_path = ''
        for k in self._known_GPI_nodes.keys():
            node = self._known_GPI_nodes.get(k)
            if top_lib is None:
                list_items.add(node.third)
            elif node.third == top_lib:
                list_items.add(node.second)
            elif node.thrd_sec == top_lib:
                list_items.add(node.name)
                if new_node_path == '':
                    new_node_path = os.path.join(node.path,
                                                 self._get_new_node_name())

        self._new_node_path = new_node_path
        if self._new_node_path == '':
            self._new_node_path_field.setText(NOPATH_MESSAGE)
        else:
            self._setQTLabelElided(self._new_node_path_field, new_node_path)

        self._new_node_list.clear()
        if top_lib is not None:
            self._new_node_list.addItem("..")

        [self._new_node_list.addItem(item) for item in list_items]

        if top_lib is None:
            idx = 0
            list_label = "GPI Libraries"
        else:
            new_label = top_lib.split('.')
            idx = len(new_label)
            list_label = u' \u2799 '.join(["GPI Libraries"] + new_label)

        self._list_label.setText(list_label)
        self._new_node_list_index = (idx, top_lib)

        if idx > 1:
            self._create_button.setEnabled(True)
        else:
            self._create_button.setDisabled(True)

    # generate the new node list window
    def generateNewNodeListWindow(self):
        # the New Node window
        self._list_win = QtWidgets.QWidget()
        self._list_win.setFixedWidth(500)
        self._new_node_list = QtWidgets.QListWidget(self._list_win)

        self._create_button = QtWidgets.QPushButton("Create Node", self._list_win)
        self._create_button.setDisabled(True)
        self._create_button.clicked.connect(self._createNewNode)
        button_layout = QtWidgets.QHBoxLayout()
        button_layout.addStretch(1)
        button_layout.addWidget(self._create_button)

        self._new_node_list.itemDoubleClicked.connect(self._listItemDoubleClicked)
        self._new_node_list.itemClicked.connect(self._listItemClicked)

        self._list_label = QtWidgets.QLabel("GPI Libraries", self._list_win)

        node_name_layout = QtWidgets.QHBoxLayout()
        new_node_name_label = QtWidgets.QLabel("Name:", self._list_win)
        self._new_node_name_field = QtWidgets.QLineEdit(self._list_win)
        self._new_node_name_field.setPlaceholderText("NewNodeName_GPI.py")
        self._new_node_name_field.textChanged.connect(self._newNodeNameEdited)
        node_name_layout.addWidget(new_node_name_label)
        node_name_layout.addWidget(self._new_node_name_field)

        new_node_path_label = QtWidgets.QLabel("Path:", self._list_win)
        self._new_node_path = ''
        self._new_node_path_field = QtWidgets.QLabel(NOPATH_MESSAGE, self._list_win)
        self._new_node_path_field.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed))
        path_layout = QtWidgets.QHBoxLayout()
        path_layout.addWidget(new_node_path_label)
        path_layout.addWidget(self._new_node_path_field)

        list_layout = QtWidgets.QVBoxLayout()
        list_layout.addWidget(self._list_label)
        list_layout.addWidget(self._new_node_list)
        list_layout.addLayout(node_name_layout)
        list_layout.addLayout(path_layout)
        list_layout.addLayout(button_layout)
        self._list_win.setLayout(list_layout)

    def scanGPIModules(self, ipath, recursion_depth=1):
        ocnt = ipath.count('/')
        for path, dn, fn in os.walk(ipath):
            # TODO: instead of checking for hidden svn dirs, just choose any hidden dir
            if (path.count('/') - ocnt <= recursion_depth) and not path.count('/.svn'):
                for fil in os.listdir(path):

                    fullpath = path+'/'+fil

                    if isGPIModFile(fullpath):

                        item = NodeCatalogItem(fullpath)
                        if Config.IMPORT_CHECK:
                            item.load()  # load check
                            if item.valid():
                                self._known_GPI_nodes.append(item)
                        else:
                            self._known_GPI_nodes.append(item)

                    elif isGPITypeFile(fullpath):

                        item = GPITYPECatalogItem(fullpath)
                        if item.valid():
                            self._known_GPI_types.append(item)

                    elif isGPINetworkFile(fullpath):

                        item = NetworkCatalogItem(fullpath)
                        if item.valid():
                            self._known_GPI_networks.append(item)

    def generateNodeSearchActions(self, txt, menu, mousemenu):

        # user query
        txt = txt.lower()

        # NODE SEARCH
        # search using txt string
        sortedMods = []
        if len(txt) > 2:  # match anywhere in name
            for node in list(self._known_GPI_nodes.values()):
                if node.name.lower().find(txt) > -1:
                    sortedMods.append(node)
        else:  # only match from start of name
            for node in list(self._known_GPI_nodes.values()):
                if node.name.lower().startswith(txt):
                    sortedMods.append(node)
        sortedMods = sorted(sortedMods, key=lambda x: x.name.lower())

        # create actions and add them to the menu
        for node in sortedMods:

            a = QtWidgets.QAction(node.name+" (" + node.thrd_sec + ")", self._parent, statusTip="Click to instantiate the \'"+str(node.name)+"\' node.")
            s = {'subsig': node}

            # The way this signal is connected, the
            # s-dict is fully copied which is required to pass the correct
            # mod name.
            a.triggered.connect(partial(self.addNodeAndCloseMouseMenu,
                                        s=s, searchmenu=menu,
                                        mousemenu=mousemenu))
            menu.addAction(a)


        # NETWORK SEARCH
        # search using txt string
        if True:
            sortedMods= []
            if len(txt) > 2:
                for net in list(self._known_GPI_networks.values()):
                    if net.name.lower().find(txt) > -1:
                        sortedMods.append(net)

            else:
                for net in list(self._known_GPI_networks.values()):
                    if net.name.lower().startswith(txt):
                        sortedMods.append(net)
            sortedMods = sorted(sortedMods, key=lambda x: x.name.lower())

            if len(sortedMods):
                menu.addSeparator()

            # create actions and add them to the menu
            for net in sortedMods:
                a = QtWidgets.QAction(net.name+" (net) (" + net.thrd_sec + ")", self._parent, statusTip="Click to instantiate the \'"+str(net.name)+"\' network.")
                s = {'sig': 'load', 'subsig': 'net', 'path': net.fullpath}
                a.triggered.connect(partial(self.addNodeAndCloseMouseMenu,
                                            s=s, searchmenu=menu,
                                            mousemenu=mousemenu))
                menu.addAction(a)



#!/usr/bin/env python

#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

# Brief: The main launcher for starting a GPI GUI session.

import sys
import os

# workaround for the Accelerate/multiprocessing bug that causes silent crashes
# when using numpy linear algebra packages on macOS
if sys.platform == 'darwin':
    os.environ["VECLIB_MAXIMUM_THREADS"] = '1'

# gpi
from gpi import QtGui, QtWidgets, QtCore, Signal
from gpi.cmd import Commands
from gpi.defines import PLOGO_PATH, ICON_PATH
from gpi.mainWindow import MainCanvas

INCLUDE_EULA=False

class Splash(QtWidgets.QSplashScreen):
    '''The splash screen that appears at GPI launch.  This contains a copy of
    the boilerplate required by Dignity Health.
    '''

    terms_accepted = Signal()

    def __init__(self, image_path):

        # find the limiting desktop dimension (w or h)
        pm = QtGui.QPixmap.fromImage(QtGui.QImage(image_path))
        g = QtWidgets.QDesktopWidget().availableGeometry()
        w = g.width()
        h = g.height()
        r = float(pm.width())/1 if pm.height() == 0 else pm.height()  # aspect ratio
        if (w <= pm.width()):
            h = int(w/r)
        if (h <= pm.height()):
            w = int(h*r)

        # the splash is almost useless below 500pts
        if w < 500:
            w = 500

        # resize the image based on new width
        if (w != g.width()) or (h != g.height()):
            pm = pm.scaledToWidth(int(w*0.8), mode=QtCore.Qt.SmoothTransformation)

        # scale subsequent parameters based on new image width
        iw = pm.width()
        ih = pm.height()

        super(Splash, self).__init__(pm)

        # use a timer instead of the EULA
        self._timer = QtCore.QTimer()
        self._timer.timeout.connect(self.terms_accepted.emit)
        self._timer.setSingleShot(True)
        if not INCLUDE_EULA:
            self._timer.start(2000)

        panel = QtWidgets.QWidget()
        pal = QtGui.QPalette(QtGui.QColor(255, 255, 255)) # white
        panel.setAutoFillBackground(True)
        panel.setPalette(pal)

        lic = '''
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
        '''
        self.lic = QtWidgets.QTextEdit(lic)
        self.lic.setReadOnly(True)

        button_title = 'Agree'
        self.wdg1 = QtWidgets.QPushButton(button_title, self)
        self.wdg1.setCheckable(False)
        self.wdg1.setFixedSize(int(iw*0.2),int(iw*0.05))
        self.wdg1.clicked[bool].connect(self.accept)

        button_title = 'Quit'
        self.wdg2 = QtWidgets.QPushButton(button_title, self)
        self.wdg2.setCheckable(False)
        self.wdg2.setFixedSize(int(iw*0.2),int(iw*0.05))
        self.wdg2.clicked[bool].connect(self.reject)

        buf = 'Click Agree to start GPI or Quit to exit.'
        new_fw = iw * 0.45
        for fw_i in range(20,0,-1):
            f = QtGui.QFont('gill sans', fw_i)
            fm = QtGui.QFontMetricsF(f)
            cfw = fm.width(buf)
            if cfw < new_fw:
                break
        f = QtGui.QFont('gill sans', fw_i)

        self.prompt = QtWidgets.QLabel(buf)
        self.prompt.setAlignment(QtCore.Qt.AlignCenter)
        self.prompt.setFont(f)

        wdgLayout = QtWidgets.QHBoxLayout()
        #wdgLayout.setContentsMargins(0, 0, 0, 0)  # no spaces around this item
        #wdgLayout.setSpacing(0)
        #wdgLayout.addSpacerItem(QtWidgets.QSpacerItem(iw/2,1,hPolicy=QtWidgets.QSizePolicy.Minimum))
        wdgLayout.addWidget(self.prompt)
        wdgLayout.addWidget(self.wdg1)
        #wdgLayout.addSpacerItem(QtWidgets.QSpacerItem(int(iw*0.01),1,hPolicy=QtWidgets.QSizePolicy.Minimum))
        wdgLayout.addWidget(self.wdg2)
        #wdgLayout.addSpacerItem(QtWidgets.QSpacerItem(1,1,hPolicy=QtWidgets.QSizePolicy.MinimumExpanding))


        # a small panel
        vbox_p = QtWidgets.QVBoxLayout()
        vbox_p.setContentsMargins(10,10,10,10)
        vbox_p.setSpacing(10)
        vbox_p.addWidget(self.lic)
        vbox_p.addLayout(wdgLayout)
        panel.setLayout(vbox_p)

        # white space | panel
        vbox = QtWidgets.QVBoxLayout()
        #vbox.setContentsMargins(0, 0, 0, int(iw*0.02))  # no spaces around this item
        vbox.setContentsMargins(0, 0, 0, 0)  # no spaces around this item
        vbox.setSpacing(0)

        vbox.addSpacerItem(QtWidgets.QSpacerItem(iw,(1-0.28)*ih,hPolicy=QtWidgets.QSizePolicy.Minimum,vPolicy=QtWidgets.QSizePolicy.Minimum))

        #vbox.addWidget(self.lic)
        vbox.addWidget(panel)
        #vbox.addLayout(wdgLayout)

        if INCLUDE_EULA:
            self.setLayout(vbox)

        self._accept = False

    def mousePressEvent(self, event):
        pass

    def accept(self):
        self.terms_accepted.emit()

    def reject(self):
        # tell Qt to quit right away.
        QtCore.QCoreApplication.instance().quit()

def launch():
    '''Starts the main application loop, parses any user config and commandline
    args.'''

    # start main application
    # for debugging force widgetcount
    #app = QtWidgets.QApplication(sys.argv+['-widgetcount'])
    app = QtWidgets.QApplication(sys.argv)
    app.setWindowIcon(QtGui.QIcon(ICON_PATH))

    # parse commandline arguments
    try:
        # PyQt4
        Commands.parse(app.argv())
    except AttributeError:
        # PyQt5
        Commands.parse(app.arguments())
    #print Commands

    # start a mainwindow widget instance
    widget = MainCanvas()

    # start splash
    # only raise in GUI mode, don't raise in cmdline mode.
    if not Commands.noGUI():

        if not Commands.noSplash():
            spl = Splash(PLOGO_PATH)

            def closeraise():
                spl.finish(widget)
                widget.show()
                widget.raise_()

            spl.terms_accepted.connect(closeraise)
            spl.show()
            spl.raise_()
            app.processEvents() # allow gui to update

        else:
            dummy = QtWidgets.QSplashScreen()
            dummy.show()
            dummy.finish(widget)
            widget.show()
            widget.raise_()

    sys.exit(app.exec_())

if __name__ == '__main__':
    launch()



#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

# Brief: A module for configuring gpi thru the ~/.gpirc file
import os
import traceback
import configparser
import glob

# gpi
from .associate import Bindings, BindCatalogItem
from gpi import VERSION
from .logger import manager
from .sysspecs import Specs

# start logger for this module
log = manager.getLogger(__name__)

# for windows
if Specs.inWindows():
    GPIRC_FILENAME = 'gpi.conf'
else:
    GPIRC_FILENAME = '.gpirc'

### ENVIRONMENT VARIABLES
if Specs.inWindows():
    USER_HOME = os.path.expanduser('~')
else:
    USER_HOME = os.environ['HOME']

USER_LIB_BASE_PATH_DEFAULT = os.path.join(USER_HOME, 'gpi')
if Specs.inWindows():
    userNameKey = 'USERNAME'
else:
    userNameKey = 'USER'
USER_LIB_PATH_DEFAULT = os.path.join(USER_LIB_BASE_PATH_DEFAULT, os.environ.get(userNameKey, 'UserNodes'))

ANACONDA_PREFIX='/opt/anaconda1anaconda2anaconda3' # is this needed?
GPI_PREFIX = os.path.dirname(os.path.realpath(__file__))
SP_PREFIX = os.path.dirname(GPI_PREFIX)

GPI_NET_PATH_DEFAULT = USER_HOME
GPI_DATA_PATH_DEFAULT = USER_HOME
GPI_FOLLOW_CWD = True

# Build the distro default to include any gpi_<name> packages in site-packages
GPI_SP_NODE_LIBS = glob.glob(os.path.join(SP_PREFIX,'gpi_*'))
GPI_LIBRARY_PATH_DEFAULT = [USER_LIB_BASE_PATH_DEFAULT,SP_PREFIX]


###############################################################################

class ConfigManager(object):
    '''An object that can load and generate the gpi config file.
    The gpi config file can potentially hold configs for:
        library paths,
        network dir path,
        data dir path,
        plugin paths,
        filetype-node associations,
        canvas window start size,
        UI style
    '''

    def __init__(self):

        # general
        self._g_import_check = True

        # root dirs for organizing gpi related files.
        self._c_networkDir = GPI_NET_PATH_DEFAULT
        self._c_dataDir = GPI_DATA_PATH_DEFAULT
        self._c_configFileName = os.path.join(os.path.expanduser('~'), GPIRC_FILENAME)

        # all the fix'ns for an initial lib
        self._c_userLibraryBasePath = os.path.expanduser(USER_LIB_BASE_PATH_DEFAULT)
        self._c_userLibraryPath = os.path.expanduser(USER_LIB_PATH_DEFAULT)
        self._c_userLibraryPath_def = os.path.join(self._c_userLibraryPath, 'default')
        self._c_userLibraryPath_def_GPI = os.path.join(self._c_userLibraryPath_def, 'GPI')
        self._c_userLibraryPath_init = os.path.join(self._c_userLibraryPath, '__init__.py')
        self._c_userLibraryPath_def_init = os.path.join(self._c_userLibraryPath_def, '__init__.py')
        self._c_userLibraryPath_def_node = os.path.join(self._c_userLibraryPath_def_GPI, 'MyNode_GPI.py')

        # env vars
        self._c_gpi_lib_path = list(GPI_LIBRARY_PATH_DEFAULT)
        self._c_gpi_follow_cwd = GPI_FOLLOW_CWD

        self._new_node_template_file = os.path.join(GPI_PREFIX, 'nodeTemplate.py')

        # make vars
        self._make_libs = []
        self._make_lib_dirs = []
        self._make_inc_dirs = []
        self._make_cflags = []

        # try to read the config file
        try:
            self.loadConfigFile()
        except:
            log.error("The config file failed to load, using defaults. "+str(traceback.format_exc()))

    def __str__(self):

        msg = ''

        # general
        msg += 'GENERAL:\n'
        for o in dir(self):
            if o.startswith('_g_'):
                msg += str(o) + ': ' + str(getattr(self, o)) + '\n'

        # path
        msg += 'PATH:\n'
        for o in dir(self):
            if o.startswith('_c_'):
                msg += str(o) + ': ' + str(getattr(self, o)) + '\n'

        # even though bindings are external print them here for convenience
        msg += 'ASSOCIATIONS:\n'
        for v in sorted([str(x) for x in list(Bindings.values())]):
            msg += str(v) + '\n'

        # makefile modifications
        msg += 'MAKE:\n'
        for o in dir(self):
            if o.startswith('_make_'):
                msg += str(o) + ': ' + str(getattr(self, o)) + '\n'
        return msg

    @property
    def IMPORT_CHECK(self):
        return self._g_import_check

    @property
    def GPI_NET_PATH(self):
        return self._c_networkDir

    @property
    def GPI_DATA_PATH(self):
        return self._c_dataDir

    @property
    def GPI_FOLLOW_CWD(self):
        return self._c_gpi_follow_cwd

    @property
    def GPI_LIBRARY_PATH(self):
        return self._c_gpi_lib_path

    @property
    def GPI_NEW_NODE_TEMPLATE_FILE(self):
        return self._new_node_template_file

    @property
    def MAKE_LIBS(self):
        return self._make_libs

    @property
    def MAKE_LIB_DIRS(self):
        return self._make_lib_dirs

    @property
    def MAKE_INC_DIRS(self):
        return self._make_inc_dirs

    @property
    def MAKE_CFLAGS(self):
        return self._make_cflags

    def generateUserLib(self):

        self.initLibDir(self._c_userLibraryBasePath)
        self.initLibDir(self._c_userLibraryPath)
        self.initLibDir(self._c_userLibraryPath_def)
        self.initLibDir(self._c_userLibraryPath_def_GPI)

        self.initLibFile(self._c_userLibraryPath_init)
        self.initLibFile(self._c_userLibraryPath_def_init)

        if os.path.exists(self._c_userLibraryPath_def_node):
            log.dialog('The user library example node: '+str(self._c_userLibraryPath_def_node) + ' already exists, skipping.')
        else:
            with open(self._c_userLibraryPath_def_node, 'w') as initfile:
                log.dialog('Writing the example node: '+str(self._c_userLibraryPath_def_node) + '')
                initfile.write(self.exampleNodeCode())

    def exampleNodeCode(self):

        header = '# GPI (v'+str(VERSION)+') auto-generated library file.\n#\n'
        filename = '# FILE: '+str(self._c_userLibraryPath_def_node)+'\n#\n'

        buf = '''# For node API examples (i.e. widgets and ports) look at the
# core.interfaces.Template node.

import gpi

class ExternalNode(gpi.NodeAPI):
    \'\'\'About text goes here...
    \'\'\'

    def initUI(self):
        # Widgets
        self.addWidget('PushButton', 'MyPushButton', toggle=True)

        # IO Ports
        self.addInPort('in1', 'NPYarray')
        self.addOutPort('out1', 'NPYarray')

        return 0

    def compute(self):

        data = self.getData('in1')

        # algorithm code...

        self.setData('out1', data)

        return 0'''

        return header+filename+buf


    def initLibDir(self, path):
        if os.path.exists(path):
            log.dialog('The user library path: '+str(path) + ' already exists, skipping.')
        else:
            log.dialog('Writing the user library path: '+str(path) + '')
            os.mkdir(path)

    def initLibFile(self, path):
        if os.path.exists(path):
            log.dialog('The user library file: '+str(path) + ' already exists, skipping.')
        else:
            log.dialog('Writing the user library file: '+str(path) + '')
            with open(path, 'w') as initfile:
                initfile.write('# GPI (v'+str(VERSION)+') auto-generated library file.\n')



    def generateConfigFile(self, overwrite=False):

        # check for existing config file
        # -force user to remove, its safer
        if self.configFileExists() and not overwrite:
            log.dialog('Config file: '+str(self.configFilePath()) + ' already exists, skipping.')
            return

        with open(self._c_configFileName, 'w') as configfile:

            # Header
            configfile.write('# GPI (v'+str(VERSION)+') configuration file.\n')
            configfile.write('# Uncomment an option to activate it.\n')

            config = configparser.RawConfigParser()

            # Makefile mods
            configfile.write('\n[GENERAL]\n')
            configfile.write('# Add nodes to the library only if they \'import\'.\n')
            configfile.write('# GPI loads faster if this check is disabled.\n')
            configfile.write('#IMPORT_CHECK = False\n')

            # PATH Section
            configfile.write('\n[PATH]\n')
            configfile.write('# Add library paths for GPI nodes.\n')
            configfile.write('# Multiple paths are delimited with a \':\'.\n')
            configfile.write('#     (e.g. [default] LIB_DIRS = ~/gpi:'+GPI_PREFIX+'/gpi/node-libs/).\n')

            configfile.write('\n# A list of directories where nodes can be found.\n')
            configfile.write('# -To enable the exercises add \''+GPI_PREFIX+'/lib/gpi/doc/Training/exercises\'.\n')
            configfile.write('#LIB_DIRS = '+ ':'.join(GPI_LIBRARY_PATH_DEFAULT) + '\n')
            configfile.write('\n# Network file browser starts in this directory.\n')
            configfile.write('#NET_DIR = '+ GPI_NET_PATH_DEFAULT + '\n')
            configfile.write('\n# Widget file browser starts in this directory.\n')
            configfile.write('#DATA_DIR = '+ GPI_DATA_PATH_DEFAULT + '\n')
            configfile.write('\n# Follow the user\'s cwd. If True, the widget and network directories\n')
            configfile.write('# will change with the user input. If False, the browsers will alwasy open\n')
            configfile.write('# to the NET_DIR and DATA_DIR.\n')
            configfile.write('#FOLLOW_CWD = '+ str(GPI_FOLLOW_CWD)+ '\n')
            #configfile.write('\n# A list of directories where plugins can be found.\n')
            #configfile.write('#PLUGIN_DIRS = '+ ':'.join(GPI_PLUGIN_PATH_DEFAULT) + '\n')

            # File-type Association Section
            configfile.write('\n[ASSOCIATIONS]\n')
            configfile.write('# Add file-type associations with nodes.\n')
            configfile.write('#  ex. (file extension, node name, widget name)\n')

            # add default associations
            cnt = 0
            for key in sorted(Bindings.keys()):
                item = Bindings.get(key)
                configfile.write('#BIND_'+str(cnt) + ' = ' + str(item.asTuple()) + '\n')
                cnt += 1

            # Makefile mods
            configfile.write('\n[MAKE]\n')
            configfile.write('# Modify the gpi-make to include new libraries, library paths,\n')
            configfile.write('# and include paths.\n')
            configfile.write('# Example: (if blas is in \'/usr\' and lapack in \'/opt/lapack\'\n')
            configfile.write('#     g++ -I /usr/include -I /opt/lapack/include -L /usr/lib -L /opt/lapack/lib \n')
            configfile.write('#          -c x.cpp -lblas -llapack -o x.so -D_MY_MACRO_=helloworld -D_ANOTHER_\n')
            configfile.write('#LIBS = blas:lapack\n')
            configfile.write('#INC_DIRS = /usr/include:/opt/lapack/include\n')
            configfile.write('#LIB_DIRS = /usr/lib:/opt/lapack/lib\n')
            configfile.write('#CFLAGS = -D_MY_MACRO_=helloworld:-D_ANOTHER_\n')

        log.dialog(str(self._c_configFileName)+' written.')

    def loadConfigFile(self):
        # load private vars from config file

        if not self.configFileExists():
            log.info("loadConfigFile(): config file " + str(self._c_configFileName) + " doesn't exist, skipping.")
            return

        config = configparser.ConfigParser()
        config.read(self._c_configFileName)

        # print parse-able info
        #for s in config.sections():
        #    log.warn(str(config.items(s)))

        # actual paths and config options
        ap = lambda x: os.path.realpath(os.path.expanduser(x))  # single dirs
        aps = lambda x: [ ap(p) for p in x.split(':') ]  # multi-dirs
        ch = config.has_option  # if config has the option...
        cg = config.get
        oh = lambda x: x in os.environ
        oe = os.environ

        if config.has_section('GENERAL'):

            parm = self.parseMultiOPTS(config, 'GENERAL', 'IMPORT_CHECK', 'GPI_IMPORT_CHECK')
            if parm:
                if parm[0].lower() == 'true':
                    self._g_import_check = True
                elif parm[0].lower() == 'false':
                    self._g_import_check = False

        # PATH section
        #   Precedence is set by this config file, then env vars, then defaults.
        if config.has_section('PATH'):

            parm = self.parseMultiOPTS(config, 'PATH', 'LIB_DIRS', 'GPI_LIBRARY_PATH')
            if parm:
                parm = self.checkDirs(parm, 'PATH::LIB_DIRS')
                self._c_gpi_lib_path = parm

            parm = self.parseMultiOPTS(config, 'PATH', 'NET_DIR', 'GPI_NET_PATH')
            if parm:
                parm = self.checkDirs(parm, 'PATH::NET_DIR')
                self._c_networkDir = parm[0]  # only single dir

            parm = self.parseMultiOPTS(config, 'PATH', 'DATA_DIR', 'GPI_DATA_PATH')
            if parm:
                parm = self.checkDirs(parm, 'PATH::DATA_DIR')
                self._c_dataDir = parm[0]  # only single dir

            parm = self.parseMultiOPTS(config, 'PATH', 'FOLLOW_CWD', 'GPI_FOLLOW_CWD')
            if parm:
                if parm[0].lower() == 'true':
                    self._c_gpi_follow_cwd = True
                elif parm[0].lower() == 'false':
                    self._c_gpi_follow_cwd = False

            # parm = self.parseMultiOPTS(config, 'PATH', 'PLUGIN_DIRS', 'GPI_PLUGIN_PATH')
            # if parm:
            #     parm = self.checkDirs(parm, 'PATH::PLUGIN_DIRS')
            #     self._c_gpi_plugin_path = parm

        # File-type Association Section
        if config.has_section('ASSOCIATIONS'):

            for item in config.items('ASSOCIATIONS'):
                t = eval(str(item[1]))
                if item[0].lower().startswith('BIND_'.lower()):
                    if len(t) != 3:
                        log.error(str(self._c_configFileName) + ': error in assignment: ' + str(item))
                        continue
                    if (type(t) is not tuple) or (type(t[0]) is not str) or (type(t[1]) is not str) or (type(t[2]) is not str):
                        log.error(str(self._c_configFileName) + ': error in assignment: ' + str(item))
                    else:
                        Bindings.append(BindCatalogItem(t))

        # Makefile Section
        if config.has_section('MAKE'):
            parm = self.parseMultiOPTS(config, 'MAKE', 'LIBS', 'GPI_MAKE_LIBS')
            if parm:
                self._make_libs = parm

            parm = self.parseMultiOPTS(config, 'MAKE', 'LIB_DIRS', 'GPI_MAKE_LIB_PATH')
            if parm:
                parm = self.checkDirs(parm, 'MAKE::LIB_DIRS')
                self._make_lib_dirs = parm

            parm = self.parseMultiOPTS(config, 'MAKE', 'INC_DIRS', 'GPI_MAKE_INC_PATH')
            if parm:
                parm = self.checkDirs(parm, 'MAKE::INC_DIRS')
                self._make_inc_dirs = parm

            parm = self.parseMultiOPTS(config, 'MAKE', 'CFLAGS', 'GPI_MAKE_CFLAGS')
            if parm:
                self._make_cflags = parm

        log.dialog(str(self._c_configFileName) + ' has been loaded.')

    def parseMultiOPTS(self, config, section, option, env_opt=None, warnOnENV=True):
        # actual paths and config options
        #ap = lambda x: os.path.realpath(os.path.expanduser(x))  # single dirs
        ap = lambda x: x
        aps = lambda x: [ ap(p) for p in x.split(':') ]  # multi-dirs
        ch = config.has_option  # if config has the option...
        cg = config.get
        oh = lambda x: x in os.environ
        oe = os.environ

        if ch(section, option):
            return aps(cg(section, option))

        elif env_opt:
            if oh(env_opt):
                if warnOnENV:
                    log.warn('Setting from user environment. - '+str(env_opt))
                return aps(oe[env_opt])

    def checkDirs(self, l, opt):
        ap = lambda x: os.path.realpath(os.path.expanduser(x))  # single dirs

        # check each dir in the list and warn on non-existing dirs
        out = []
        for d in l:
            de = ap(d)  # expand paths
            out.append(de)
            if not os.path.isdir(de):
                log.warn('User Config: \''+str(opt)+'\': \''+str(d)+'\' is not a directory.')
        return out

    def configFileExists(self):
        return os.path.isfile(self._c_configFileName)

    def configFilePath(self):
        return self._c_configFileName

    def userLibPath(self):
        return self._c_userLibraryPath


# activate this upon first import
Config = ConfigManager()

#print Config



#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

# Brief: Update utility, can be called directly from gpi or run as a separate
#        program.

import os
import re
import sys
import json
import time
import subprocess

from gpi import QtGui, QtWidgets, QtCore, Signal
from .widgets import TextBox
from .runnable import ExecRunnable, Runnable
from .sysspecs import Specs

# get the anaconda path to ensure that THIS installation is being updated
ANACONDA_PREFIX = '/opt/anaconda1anaconda2anaconda3' # ANACONDA
if ANACONDA_PREFIX == '/opt/'+'anaconda1anaconda2anaconda3':
    # get the path from the user env
    if Specs.inWindows():
        ANACONDA_PREFIX = os.path.dirname(subprocess.check_output('where conda', shell=True).decode('latin1').strip())
    else:
        ANACONDA_PREFIX = os.path.dirname(subprocess.check_output('which conda', shell=True).decode('latin1').strip())
    ANACONDA_PREFIX = os.path.dirname(ANACONDA_PREFIX) # strip off the 'bin'

class JSONStreamLoads(object):
    ''' Load multiple json objects from string.
    Returns loaded objects in a list.
    '''

    def __init__(self, in_str, linefeed=True):

        if type(in_str) == str:
            self._buffer = in_str
        else:
            raise TypeError("JSONStreamLoads(): input must be of type \'str\'.")

        self._load = self.loadAll()

    def load(self):
        return self._load

    def loadAll(self):
        return json.loads(self._buffer)


# use conda to update to the latest package
class CondaUpdater(QtCore.QObject):
    '''Handles the underlying communication with the 'conda' program.
    '''

    pdone = Signal(int)
    message = Signal(str)
    failed = Signal(str)
    getStatus_done = Signal()
    updateAllPkgs_done = Signal()

    def __init__(self, conda_prefix=ANACONDA_PREFIX, dry_run=False):
        super().__init__()
        self._dry_run = dry_run
        self._conda_prefix = conda_prefix
        self._packages = ['gpi', 'gpi_core']
        # DDB - No docs for now on c-f, restore later
        # self._packages = ['gpi', 'gpi_core', 'gpi-docs']

        self._packages_for_installation = []
        self._packages_for_update = []

        self._current_versions = {}
        self._latest_versions = {}


    def _status_pdone(self, pct, cr=False):
        end = ''
        if pct == 100:
            end = '\n'
        msg = 'Searching for package updates: '+str(pct)+'%'
        print('\t'+msg+'\r', end=end)
        self.pdone.emit(pct)
        self.message.emit('Searching for package updates...')

    def getStatus(self):
        try:
            self._getStatus()
            self.getStatus_done.emit()
        except:
            self.failed.emit('Failed to fetch updates.')

    def _getStatus(self):

        self.checkConda()

        # total divisions are len(self._packages)*3
        pdone = 0
        divs = len(self._packages)*3 + 1
        step = int(100/divs)
        self._status_pdone(1)

        # Check for the current installed versions
        for pkg in self._packages:
            self._current_versions[pkg] = self.getInstalledPkgVersion(pkg)
            pdone += step
            self._status_pdone(pdone)

        # Check for the latest versions online
        for pkg in self._packages:
            if self._current_versions[pkg] is None:
                self._latest_versions[pkg] = self.updatePkg(pkg, dry_run=True, install=True)
            else:
                self._latest_versions[pkg] = self.updatePkg(pkg, dry_run=True)
            pdone += step
            self._status_pdone(pdone)

        # Sort targets into 'install' or 'update'
        for pkg in self._packages:
            # updates - if there is both an installed version and new version
            if (self._latest_versions[pkg] is not None) and \
                (self._current_versions[pkg] is not None):
                self._packages_for_update.append(pkg)
            # installs - if there is no installed version, the latest will be
            #            whatever is available.
            if (self._latest_versions[pkg] is not None) and \
                (self._current_versions[pkg] is None):
                self._packages_for_installation.append(pkg)
            pdone += step
            self._status_pdone(pdone)

        self._status_pdone(100)

    def __str__(self):
        msg = ''
        tab = '&nbsp;&nbsp;&nbsp;&nbsp;'

        # updates
        if len(self._packages_for_update):
            msg += 'The following packages will be updated:<br><br>'
            for pkg in self._packages_for_update:
                o = self._current_versions[pkg]
                n = self._latest_versions[pkg]
                msg += pkg + '<br>'
                msg += tab+str(o) + '&nbsp; &#10154; &nbsp;' + str(n) + '<br>'

        # installs
        if len(self._packages_for_installation):
            if msg != '':
                msg += '<br><br>'
            msg += 'The following packages will be installed:<br><br>'
            for pkg in self._packages_for_installation:
                n = self._latest_versions[pkg]
                msg += pkg + '<br>'
                msg += tab+str(n) + '<br>'

        if self.numberOfUpdates():
            msg += '<br><br>GPI will be <b>automatically restarted</b> after updating.' \
                 + '  Please make sure your networks are saved before proceeding.'

        if msg == '':
            msg = 'GPI is up to date.'

        return msg

    def statusMessage(self):
        return str(self)

    def checkConda(self):
        cmd_unix = 'conda --version >/dev/null 2>&1'
        cmd_win = 'conda --version > NUL' 
        try:
            if Specs.inWindows():
                subprocess.check_output(cmd_win, shell=True)
            else:
                subprocess.check_output(cmd_unix, shell=True)
        except subprocess.CalledProcessError as e:
            print('Failed to execute conda, aborting...')
            print(e.cmd, e.output)
            raise
        except:
            print('\'conda\' failed to execute, aborting...')
            print(cmd)
            raise

    def getInstalledPkgVersion(self, name):
        cmd = 'conda list --json'
        try:
            output = subprocess.check_output(cmd, shell=True).decode('utf8')
            conda = JSONStreamLoads(output).load()
            for pkg in conda:
                pkg_str = json.dumps(pkg)
                m = re.search(name+'-([0-9]+\.[0-9]+\.[0-9]+)-[^"]*', pkg_str)
                if m:
                    return m[1]
        except:
            print('Failed to retrieve installed package information on '+name+', skipping...')
            print(cmd)
            raise

    def _updateAllPkgs_pdone(self, pct, cr=False):
        end = ''
        if pct == 100:
            end = '\n'
        msg = 'Updating packages: '+str(pct)+'%'
        print('\t'+msg+'\r', end=end)
        self.pdone.emit(pct)

    def numberOfUpdates(self):
        return len(self._packages_for_installation) + len(self._packages_for_update)

    def updateAllPkgs(self):
        try:
            self._updateAllPkgs()
            self.updateAllPkgs_done.emit()
        except:
            self.failed.emit('Failed to install updates.')

    def _updateAllPkgs(self):
        if self._dry_run:
            self.message.emit('Package updates complete. Relaunching...')
            return

        # total divisions are the installation list plus the update list
        pdone = 0
        divs = self.numberOfUpdates() + 1
        step = int(100/divs)
        self._updateAllPkgs_pdone(1)

        tab = '&nbsp;&nbsp;&nbsp;&nbsp;'
        message_hdr = 'Updating packages...<br>'+tab

        # Install or update all the packages that have been determined.
        for pkg in self._packages_for_installation:
            # if there is no package (due to user changes) then install it
            self.updatePkg(pkg, install=True)
            pdone += step
            self._updateAllPkgs_pdone(pdone)
            self.message.emit(message_hdr+pkg)
        for pkg in self._packages_for_update:
            # if there is a latest version then update
            self.updatePkg(pkg)
            pdone += step
            self._updateAllPkgs_pdone(pdone)
            self.message.emit(message_hdr+pkg)

        self._updateAllPkgs_pdone(100)
        self.message.emit('Package updates complete. Relaunching...')

    def updatePkg(self, name, dry_run=False, install=False):
        # Updates to the latest package and returns the package string.
        #   -dry_run will just return the latest package string.
        #   -install will install the package if its not currently installed.

        conda_sub = 'update'
        if install: conda_sub = 'install'
        dry_cmd = ''
        if dry_run: dry_cmd = '--dry-run --no-deps'
        cmd = 'conda '+conda_sub+' '+name+' -y --json '+dry_cmd

        try:
            output = subprocess.check_output(cmd, shell=True).decode('utf8')
            if dry_run:
                conda = JSONStreamLoads(output).load()

                if conda['success']:
                    if 'message' in conda: # if we're up to date
                        return
                    for pkg in conda['actions']['LINK']:
                        pkg_str = pkg['dist_name']
                        if pkg_str.startswith(name):
                            m = re.search('-([0-9]+\.[0-9]+\.[0-9]+)-', pkg_str)
                            return m[1]
                else:
                    raise RuntimeError('conda returned a failure status.')
            else:
                m = re.search('\"success\": true', output)
                if m[0] is not None:
                    print("Successfully updated ",name)
                else:
                    print("Update unsuccessful, even though a newer package was found.")
                    print("This is likely due to a dependency problem. Please create an")
                    print("issue at https://github.com/conda-forge/gpi-feedstock ")
        except subprocess.CalledProcessError as e:
            print('Failed to update to new package. Please try again.')
            print('If the problem persists, create an issue at')
            print('https://github.com/conda-forge/gpi-feedstock')
            print(e.cmd, e.output)
            raise
        except:
            print('Failed to retrieve package update information.')
            print('This may be due to a network or package manager issue.')
            print('If the problem persists, create an issue at')
            print('https://github.com/conda-forge/gpi-feedstock')
            print(cmd)
            raise

class UpdateWindow(QtWidgets.QWidget):
    '''A simple UI to display the GPI update process as its happening.
    '''

    _startGetStatus = Signal()

    def __init__(self, dry_run=False):
        super().__init__()

        self._updater = CondaUpdater(dry_run=dry_run)
        self._updater.getStatus_done.connect(self.showStatus)
        self._updater.getStatus_done.connect(self.showOKorUpdateButton)
        self._updater.failed.connect(self.initFailureMode)
        self._updater.updateAllPkgs_done.connect(self.relaunchProc)

        style = '''
            QProgressBar {
                background-color: rgb(226,226,226);
                border: 1px solid rgb(222,222,222);
                border-radius: 2px;
                text-align: center;
            }

            QProgressBar::chunk {
                background-color: #0099FF;
                height: 15px;
                width: 1px;
            }
        '''
        self._pbar = QtWidgets.QProgressBar(self)
        self._pbar.setStyleSheet(style)
        self._updater.pdone.connect(self.pdone)

        self._txtbox = TextBox('')
        self._txtbox.wdg.setTextFormat(QtCore.Qt.RichText)
        self._txtbox.set_wordwrap(True)
        self._txtbox.set_openExternalLinks(True)
        self._updater.message.connect(self._txtbox.set_val)
        self._txtbox.set_val('Checking for updates...')

        self._okButton = QtWidgets.QPushButton("OK")
        self._okButton.setVisible(False)
        self._cancelButton = QtWidgets.QPushButton("Cancel")
        self._cancelButton.clicked.connect(self.close)

        hbox = QtWidgets.QHBoxLayout()
        hbox.addStretch(1)
        hbox.addWidget(self._okButton)
        hbox.addWidget(self._cancelButton)

        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(self._txtbox, 1)
        vbox.addWidget(self._pbar)
        vbox.addLayout(hbox)

        self.setLayout(vbox)

        self.setGeometry(300, 300, 400, 350)
        self.setWindowTitle('GPI Update')
        self.show()
        self.raise_()

        ExecRunnable(Runnable(self._updater.getStatus))

    def installUpdates(self):
        self._okButton.setVisible(False)
        ExecRunnable(Runnable(self._updater.updateAllPkgs))

    def pdone(self, pct):
        self._pbar.setValue(pct)
        if pct < 100:
            self._pbar.setVisible(True)
        else:
            self._pbar.setVisible(False)
            self._pbar.setValue(0)

    def showStatus(self):
        self._txtbox.set_val(self._updater.statusMessage())

    def initFailureMode(self, msg):
        self._pbar.setVisible(False)
        self._txtbox.set_val(msg)
        self._okButton.setVisible(False)
        self._cancelButton.setText('Close')

    def showOKorUpdateButton(self):
        if self._updater.numberOfUpdates():
            self._okButton.setText('Update && Relaunch')
            self._okButton.setVisible(True)
            self._okButton.clicked.connect(self.installUpdates)
        else:
            self._okButton.setVisible(False)
            self._cancelButton.setText('Close')

    def relaunchProc(self):
        try:
            self.update()
            QtWidgets.QApplication.processEvents() # allow gui to update
            time.sleep(5)
            args = sys.argv[:]
            args.insert(0, sys.executable)
            os.execv(sys.executable, args)
        except:
            self.initFailureMode('Failed to relaunch GPI.')

# For running as a separate application.
def update():
    app = QtWidgets.QApplication(sys.argv)
    win = UpdateWindow()
    sys.exit(app.exec_())



#!/usr/bin/env python

#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#    The code in this file was modifed/derived from the topsort.py with the
#    following license:
#
#       Permission is hereby granted to copy, modify and use the above source
#       code for any purpose as long as the following comment line is included
#       with it:
#
#           Original topological sort code written by Ofer Faigon
#           (www.bitformation.com) and used with permission

def topological_sort(items, partial_order):
    """Perform topological sort.
       items is a list of items to be sorted.
       partial_order is a list of pairs. If pair (a,b) is in it, it means
       that item a should appear before item b.
       Returns a list of the items in one of the possible orders, or None
       if partial_order contains a loop.
    """

    def add_node(graph, node):
        """Add a node to the graph if not already exists."""
        if node not in graph:
            graph[node] = [0] # 0 = number of arcs coming into this node.

    def add_arc(graph, fromnode, tonode):
        """Add an arc to a graph. Can create multiple arcs.
           The end nodes must already exist."""
        graph[fromnode].append(tonode)
        # Update the count of incoming arcs in tonode.
        graph[tonode][0] = graph[tonode][0] + 1

    # step 1 - create a directed graph with an arc a->b for each input
    # pair (a,b).
    # The graph is represented by a dictionary. The dictionary contains
    # a pair item:list for each node in the graph. /item/ is the value
    # of the node. /list/'s 1st item is the count of incoming arcs, and
    # the rest are the destinations of the outgoing arcs. For example:
    #           {'a':[0,'b','c'], 'b':[1], 'c':[1]}
    # represents the graph:   c <-- a --> b
    # The graph may contain loops and multiple arcs.
    # Note that our representation does not contain reference loops to
    # cause GC problems even when the represented graph contains loops,
    # because we keep the node names rather than references to the nodes.
    graph = {}
    for v in items:
        add_node(graph, v)
    for a,b in partial_order:
        add_arc(graph, a, b)

    # Step 2 - find all roots (nodes with zero incoming arcs).
    roots = [node for (node,nodeinfo) in list(graph.items()) if nodeinfo[0] == 0]

    # step 3 - repeatedly emit a root and remove it from the graph. Removing
    # a node may convert some of the node's direct children into roots.
    # Whenever that happens, we append the new roots to the list of
    # current roots.
    sorted = []
    while len(roots) != 0:
        # If len(roots) is always 1 when we get here, it means that
        # the input describes a complete ordering and there is only
        # one possible output.
        # When len(roots) > 1, we can choose any root to send to the
        # output; this freedom represents the multiple complete orderings
        # that satisfy the input restrictions. We arbitrarily take one of
        # the roots using pop(). Note that for the algorithm to be efficient,
        # this operation must be done in O(1) time.
        root = roots.pop()
        sorted.append(root)
        for child in graph[root][1:]:
            graph[child][0] = graph[child][0] - 1
            if graph[child][0] == 0:
                roots.append(child)
        del graph[root]
    if len(list(graph.items())) != 0:
        # There is a loop in the input.
        return None
    return sorted

def topsort(connection_list):
    """NRZ:
       The connection list is a list of pairs indicating flow direction through node ordering.
        -we only care about nodes with connections
            -so only itemize those nodes
    """
    # remove redundant connections
    c = set(connection_list)
    
    # list connected nodes
    l = set([ i for i,j in c ] + [ j for i,j in c ])

    return topological_sort(l,c)


if __name__ == '__main__':

    print("test topsort")
    print((topological_sort([1,2,3], [(1,2),(1,3),(3,2)]))) # --> [1,3,2]
    print((topological_sort([1,2], [(2,1),(2,1)]))) # --> [2,1]
    print((topological_sort([1,2], [(1,2),(2,1)]))) # --> None
    print((topological_sort([0,1,2], [(0,1),(1,2),(2,1)]))) # --> None


    # hashable type
    x = type
    y = int
    z = float
    print((topological_sort([x,y,z], [(x,y),(x,z),(z,y)]))) # --> [1,3,2]

    print((topsort( [(x,y),(x,z),(z,y)] )))

    print("cyclic and acylic graphs")
    print((topsort( [(1,2),(1,3),(3,2),(5,6),(5,7),(7,6),(7,5)] )))

    print("multiple graphs")
    print((topsort( [(1,2),(1,3),(3,2),(5,6),(5,7),(7,6)] )))







#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.


from gpi import QtCore, QtGui, QtWidgets

# gpi
from .defines import PortTYPE, InPortTYPE, OutPortTYPE, REQUIRED, OPTIONAL
from .defines import GPI_PORT_EVENT, stw
from .defines import getKeyboardModifiers, printMouseEvent
from .logger import manager

# start logger for this module
log = manager.getLogger(__name__)


class Port(QtWidgets.QGraphicsItem):
    '''The base-class of the Node InPorts and OutPorts. This is responsible for
    Tool tips, serialization, drawing, painting, and type matching for
    connectivity.
    '''
    Type = PortTYPE

    def __init__(self, nodeWidget, CanvasBackend, portTitle, portNum, intype=None, dtype=None, ndim=None, menuWidget=None):
        super(Port, self).__init__()

        self._id = None
        self.setID()

        self.node = nodeWidget
        self.menuWidget = menuWidget  # a reference to a module menu widget
        self.graph = CanvasBackend
        self.edgeList = []
        self.newPos = QtCore.QPointF()

        self.setFlag(QtWidgets.QGraphicsItem.ItemSendsGeometryChanges)
        self.setCacheMode(QtWidgets.QGraphicsItem.DeviceCoordinateCache)
        self.setZValue(3)

        ## TYPE info
        self._GPIType = intype  # basic data type
        # for NPY arrays
        self._dtype = dtype  # array data type
        self._ndim = ndim
        self._obligation = None  # inport use only
        # a link to the data
        self._data = None

        # position
        self.portNum = portNum
        self.portTitle = portTitle
        if portTitle is None:
            self.portTitle = str(portNum)

        # port shape
        self.largenessFact = 3.0
        self.pointsCoord = [[0.0, 0.0], [7.0, 0.0], [3.5, 5]]
        self.setTransformOriginPoint(3.5, 5)
        if isinstance(self, OutPort):
            self.setTransformOriginPoint(3.5, 0)
        # self.portShape = trianglePolygon = QtGui.QPolygonF()
        self.portShape = QtGui.QPolygonF()
        for i in self.pointsCoord:
            self.portShape.append(QtCore.QPointF(i[0], i[1]))

        # box
        # self.pointsCoord_canvasConnect = [[0.0, 0.0], [7.0, 0.0], [7.0, 5], [0.0, 5]]
        # bowtie
        #self.pointsCoord_canvasConnect = [[0.0, 0.0], [7.0, 0.0], [3.5, 2.5], [0.0, 5.0], [7.0, 5.0], [3.5, 2.5]]
        #self.portShape_canvasConnect = QtGui.QPolygonF()
        #for i in self.pointsCoord_canvasConnect:
        #    self.portShape_canvasConnect.append(QtCore.QPointF(i[0], i[1]))

        # box
        # self.pointsCoord_canvasConnect = [[0.0, 0.0], [7.0, 0.0], [7.0, 5], [0.0, 5]]
        # bowtie
        self.pointsCoord_memSave = [[0.0, 0.0], [7.0, 0.0], [3.5, 2.5], [0.0, 5.0], [7.0, 5.0], [3.5, 2.5]]
        self.portShape_memSave = QtGui.QPolygonF()
        for i in self.pointsCoord_memSave:
            self.portShape_memSave.append(QtCore.QPointF(i[0], i[1]))

        self.setCursor(QtCore.Qt.CrossCursor)

        self.setAcceptHoverEvents(True)
        self._beingHovered = False
        self.updateToolTip()

        # src or sink
        self._canvasConnect = False

        # save memory
        self._savemem = False

    def triggerHoverLeaveEvent(self):
        e = QtCore.QEvent(QtCore.QEvent.GraphicsSceneHoverLeave)
        self.hoverLeaveEvent(e)

    def BeingHovered(self):
        return self._beingHovered

    def allowsCyclicConn(self):
        '''Abstract function
        '''
        return False

    def resetPos(self):
        if isinstance(self, InPort):
            self.setPos(-8 + 8 * self.portNum, -12)
            self.updateEdges()

        if isinstance(self, OutPort):
            h = self.getNode().getOutPortVOffset()
            self.setPos(-8 + 8 * self.portNum, h)
            self.updateEdges()

    def setPosByPortNum(self, portNum):
        if isinstance(self, InPort):
            self.setPos(-8 + 8 * portNum, -12)
            self.updateEdges()

        if isinstance(self, OutPort):
            h = self.getNode().getOutPortVOffset()
            self.setPos(-8 + 8 * portNum, h)
            self.updateEdges()

    def setMemSaver(self, val):
        self._savemem = val

    def isMemSaver(self):
        return self._savemem

    def toggleMemSaver(self):
        if self._savemem:
            self._savemem= False
        else:
            self._savemem= True

    def setCanvasConnect(self, val):
        self._canvasConnect= val

    def isCanvasConnected(self):
        return self._canvasConnect

    def toggleCanvasConnect(self):
        if self._canvasConnect:
            self._canvasConnect = False
        else:
            self._canvasConnect = True

    def setID(self, value=None):
        if value is None:
            self._id = id(self)  # this will always be unique
        else:
            self._id = value

    def getID(self):
        return self._id

    def isWidgetPort(self):
        if self.menuWidget is None:
            return False
        else:
            return True

    def GPIType(self):
        return self._GPIType

    def updateToolTip(self):
        # PORTAUTH
        oblBuf = ''
        if isinstance(self, InPort):
            if self.isREQUIRED():
                oblBuf = '(REQUIRED)'
            else:
                oblBuf = '(OPTIONAL)'

            if self.allowsCyclicConn():
                oblBuf += ' Cyclic'

        tip = "\'"+str(self.portTitle)+'\': '+oblBuf+'\n'
        tip += self._GPIType.toolTip_Port()
        tip += '\n'

        if isinstance(self, OutPort):
            tip += '____________________\n'
            tip += 'Data Info: \n'
            tip += self._GPIType.toolTip_Data(self.data)

        self.setToolTip(tip)

    def scaleUp(self):
        self.setScale(self.largenessFact)
        self.update()

    def resetScale(self):
        self.setScale(1)
        self.update()

    def findOppositePorts(self):

        ports = []

        # for outports, find all appropriate inports
        if isinstance(self, OutPort):
            for item in list(self.graph.scene().items()):
                if isinstance(item, InPort):
                    ports.append(item)

        # vica-versa for all inports
        if isinstance(self, InPort):
            for item in list(self.graph.scene().items()):
                if isinstance(item, OutPort):
                    ports.append(item)

        return ports

    def getName(self):
        return self.portTitle

    def getSettings(self):  # PORT SETTINGS
        # get the settings required to re-instantiate this port
        # PORTAUTH
        s = {}

        # Possibly not needed:
        s['id'] = self.getID()
        s['porttitle'] = self.portTitle
        s['porttype'] = self.porttype()

        # data type
        # s['intype']      = self._type
        # s['dtype']       = self._dtype
        # s['ndim']        = self._ndim

        s['obligation'] = self._obligation
        s['portnum'] = self.portNum
        if self.isWidgetPort():
            s['widgetTitle'] = self.menuWidget.getTitle()
        else:
            s['widgetTitle'] = None

        # In loadNetwork() this is the only info used
        # -the rest is instantiated from ExternalNode()
        # derived class.
        s['connections'] = []
        for edge in self.edgeList:
            s['connections'].append(edge.getCoords())
        return s

    def getNodeID(self):
        return self.node.getID()

    def hoverEnterEvent(self, event):

        # Upstream forced hover
        if isinstance(self, InPort):
            for edge in self.edgeList:
                edge.sourcePort().hoverEnterEvent(event)

        self._beingHovered = True
        self.setScale(self.largenessFact)
        self.node.setZValue(101)
        self.update()
        for edge in self.edgeList:
            edge.adjust()
            edge.update()
        self.node.graph.viewAndSceneForcedUpdate()

    def hoverLeaveEvent(self, event):

        # Upstream forced hover
        if isinstance(self, InPort):
            for edge in self.edgeList:
                edge.sourcePort().hoverLeaveEvent(event)

        self._beingHovered = False
        self.setScale(1.0)
        self.node.setZValue(2)
        self.update()
        for edge in self.edgeList:
            edge.adjust()
            edge.update()
        self.node.graph.viewAndSceneForcedUpdate()

    def type(self):
        return Port.Type

    def porttype(self):
        return Port.Type

    def addEdge(self, edge):
        self.edgeList.append(edge)
        edge.adjust()

    def detachEdge(self, edge):
        '''This is a little misleading, it only pops the edge from a port\'s edgelist'''
        self.triggerHoverLeaveEvent()
        for i in range(len(self.edgeList)):
            if self.edgeList[i] == edge:
                return self.edgeList.pop(i)

    def edges(self):  # PORT
        return self.edgeList

    def edgeCount(self):
        return len(self.edgeList)

    # all connected nodes and thier connetion directions
    def getConnectionTuples(self):
        '''Source and dest only apply to the
        order in which the user connected the edge.
        The DAG order is determined by port direction.
        -append( (src -> sink) )
        '''
        c = []
        for edge in self.edgeList:
            if isinstance(edge.sourcePort(), OutPort):
                c.append((edge.sourceNode(), edge.destNode()))
            else:
                c.append((edge.destNode(), edge.sourceNode()))
        return c

    def getNonCyclicConnectionTuples(self):
        c = []
        for edge in self.edgeList:
            if not edge.isCyclicConnection():
                if isinstance(edge.sourcePort(), OutPort):
                    c.append((edge.sourceNode(), edge.destNode()))
                else:
                    c.append((edge.destNode(), edge.sourceNode()))
        return c

    def getDownstreamNodes(self):  # PORT
        '''Get downstream nodes connected to this port.'''
        c = []
        for edge in self.edgeList:
            edge.update()  # update edge-text
            if edge.sourcePort() is self:
                # c[edge.destNode()] = edge.destPort().portTitle
                c.append([edge.destNode(), edge.destPort().portTitle])
                # print edge.destNode().name
        return c

    def getNode(self):
        return self.node

    def boundingRect(self):
        adjust = 2.0
        maxx = max([i for i, j in self.pointsCoord])
        maxy = max([j for i, j in self.pointsCoord])
        return QtCore.QRectF((0 - adjust), (0 - adjust), (maxx + adjust), (maxy + adjust))

    def shape(self):
        path = QtGui.QPainterPath()
        # path.addEllipse(0, 0, 5, 5)
        path.addPolygon(self.portShape)
        self.update()
        return path

    def paint(self, painter, option, widget):  # PORT
        # choose module color
        gradient = QtGui.QRadialGradient(-1, -1, 10)
        if option.state & QtWidgets.QStyle.State_Sunken:
            gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.red).lighter(150))
            gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.darkRed).lighter(150))
        # elif self._beingHovered:
        #    gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.darkRed).lighter(150))
        #    gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.red).lighter(150))
        elif isinstance(self, InPort):
            # if self.menuWidget:
            if self.isREQUIRED():
                # gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.yellow).lighter(200))
                # gradient.setColorAt(1,
                # QtGui.QColor(QtCore.Qt.darkYellow).lighter(200))
                gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.blue).lighter(200))
                gradient.setColorAt(1, QtGui.QColor(
                    QtCore.Qt.darkBlue).lighter(200))
                # gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.gray).lighter(300))
                # gradient.setColorAt(0,
                # QtGui.QColor(QtCore.Qt.darkGray).lighter(150))
            else:
                gradient.setColorAt(0, QtGui.QColor(
                    QtCore.Qt.green).lighter(200))
                gradient.setColorAt(1, QtGui.QColor(
                    QtCore.Qt.darkGreen).lighter(150))
            #    gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.gray).lighter(150))
            # gradient.setColorAt(1,
            # QtGui.QColor(QtCore.Qt.darkGray).lighter(100))
        elif isinstance(self, OutPort):
            # if self.menuWidget:
                # orange=QtGui.QColor().fromRgbF(1.,0.5,0.)
                # gradient.setColorAt(0, orange.lighter(200))
                # gradient.setColorAt(1, orange.lighter(100))
                # gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.blue).lighter(300))
                # gradient.setColorAt(1,
                # QtGui.QColor(QtCore.Qt.darkBlue).lighter(300))
                if self.dataIsNone():
                    gradient.setColorAt(1, QtGui.QColor(
                        QtCore.Qt.red).lighter(300))
                    gradient.setColorAt(0, QtGui.QColor(
                        QtCore.Qt.darkRed).lighter(150))
                elif self.dataHasChanged():
                    # gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.gray).lighter(200))
                    # gradient.setColorAt(0,
                    # QtGui.QColor(QtCore.Qt.darkGray).lighter(100))
                    gradient.setColorAt(1, QtGui.QColor(
                        QtCore.Qt.blue).lighter(200))
                    gradient.setColorAt(0, QtGui.QColor(
                        QtCore.Qt.darkBlue).lighter(200))
                else:
                    gradient.setColorAt(1, QtGui.QColor(
                        QtCore.Qt.yellow).lighter(200))
                    gradient.setColorAt(0, QtGui.QColor(
                        QtCore.Qt.darkYellow).lighter(150))
            # else:
                # gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.blue).lighter(175))
                # gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.darkBlue).lighter(175))
                # gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.gray).lighter(150))
                # gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.darkGray).lighter(100))
            #    if self.dataIsNone():
            #        gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.red).lighter(150))
            #        gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.darkRed).lighter(100))
            #    elif self.dataHasChanged():
            #        gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.gray).lighter(150))
            #        gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.darkGray).lighter(100))
            #    else:
            #        gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.yellow).lighter(200))
            # gradient.setColorAt(1,
            # QtGui.QColor(QtCore.Qt.darkYellow).lighter(150))

        # draw module box (apply color)
        painter.setBrush(QtGui.QBrush(gradient))
        #painter.setPen(QtGui.QPen(QtCore.Qt.black, 0))
        fade = QtGui.QColor(QtCore.Qt.black)
        fade.setAlpha(50)
        painter.setPen(QtGui.QPen(fade, 0))

        # if isinstance(self,InPort):
        #    if self.isREQUIRED():
        #        painter.setPen(QtGui.QPen(QtCore.Qt.red, 0))
        #    else:
        #        painter.setPen(QtGui.QPen(QtCore.Qt.black, 0))
        #        #painter.setPen(QtGui.QPen(QtCore.Qt.yellow, 0))

        if self.isMemSaver():
            painter.drawPolygon(self.portShape_memSave)
        else:
            if self.menuWidget:
                painter.drawEllipse(1, 0, 5, 5)
            else:
                painter.drawPolygon(self.portShape)

    def updateEdges(self):
        for edge in self.edgeList:
            edge.adjust()

    def mousePressEvent(self, event):  # PORT
        printMouseEvent(self, event)
        modifiers = getKeyboardModifiers()

        self.update()  # update port color/size/etc
        modmidbutton_event = (event.button() == QtCore.Qt.LeftButton
                              and modifiers == QtCore.Qt.AltModifier)
        if event.button() == QtCore.Qt.MidButton or modmidbutton_event:
            event.accept()
        elif event.button() == QtCore.Qt.LeftButton:
            event.accept()
        elif event.button() == QtCore.Qt.RightButton:
            event.accept()
        else:
            event.ignore()
        super(Port, self).mousePressEvent(event)

    def mouseReleaseEvent(self, event):  # PORT
        printMouseEvent(self, event)

        self.update()  # update port color/size/etc
        if event.button() == QtCore.Qt.MidButton:
            event.accept()
        elif event.button() == QtCore.Qt.LeftButton:
            event.accept()
        elif event.button() == QtCore.Qt.RightButton:
            event.accept()
        else:
            event.ignore()
        super(Port, self).mouseReleaseEvent(event)

    def advance(self):
        if self.newPos == self.pos():
            return False

        self.setPos(self.newPos)
        return True

    def checkDataType(self, indata):
        # PORTAUTH
        # This checks the inbound data type against the porttype while a
        # connection is being made or when new data is passed downstream.
        return self._GPIType.matchesData(indata)

    def checkTypeAgainstPort(self, uport):
        # PORTAUTH
        # This check verfies a connection between two ports
        # when the user is drawing a connection or when a port
        # is clicked and other connectable ports are highlighted.
        if uport:
            return self._GPIType.matchesType(uport._GPIType)

        # port doesn't exist, not sure why this is being tested.
        return False


class InPort(Port):
    '''Defines the specific behavior for connecting to inports i.e. how to 
    check for upstream data types, and obligation.
    '''

    PortType = InPortTYPE

    def __init__(self, nodeWidget, CanvasBackend, title, portNum, intype=None, dtype=None, ndim=None, obligation=REQUIRED, menuWidget=None, cyclic=False):

        self._cyclic = cyclic

        super(InPort, self).__init__(nodeWidget, CanvasBackend,
                                     title, portNum, intype, dtype, ndim, menuWidget)
        # REQUIRED or OPTIONAL type
        self._obligation = obligation

    def edge(self):
        '''InPorts only have one edge'''
        if len(self.edgeList):
            return self.edgeList[0]

    def allowsCyclicConn(self):
        return self._cyclic

    # matching outports
    def findMatchingOutPorts(self, ports=None):
        if ports == None: ports = self.findOppositePorts()
        matching_ports = []

        # just ignore if its already connected
        if len(self.edges()):
            return matching_ports

        if len(ports):
            for port in ports:
                # skip ports on the same node
                if port.node == self.node:
                    continue
                # check data if available
                if port.data is not None:
                    if self.checkDataType(port.data):
                        matching_ports.append(port)
                # if not, just check against port spec
                else:
                    # call checker from inport since they are
                    # the limiting factor
                    if self.checkTypeAgainstPort(port):
                        matching_ports.append(port)

        return matching_ports

    def incomingDataTypeMatches(self):
        indata = self.getUpstreamData()
        if indata is not None:
            return self.checkDataType(indata)
        else:
            return False

    def checkUpstreamPortType(self):
        uport = self.getUpstreamPort()
        if uport.data is not None:
            return self.checkDataType(uport.data)
        else:
            return self.checkTypeAgainstPort(uport)

    def type(self):  # generic object type
        return InPort.Type

    def porttype(self):
        return InPort.PortType

    def getUpstreamData(self):
        # first check if there is any data
        if len(self.edgeList) > 0:
            return self.edgeList[0].sourcePort().data
        else:
            return None

    def getUpstreamPort(self):
        # first check if there is any data
        if len(self.edgeList) > 0:
            return self.edgeList[0].sourcePort()
        else:
            return None

    def getUpstreamNode(self):
        '''Return node ref in a list.
        '''
        c = []
        if self.getUpstreamPort():
            c.append(self.getUpstreamPort().getNode())
        return c

    def setREQUIRED(self):
        self._obligation = REQUIRED
        log.debug("set REQUIRED")
        return self

    def setOPTIONAL(self):
        self._obligation = OPTIONAL
        log.debug("set OPTIONAL")
        return self

    def isOPTIONAL(self):
        return self._obligation == OPTIONAL

    def isREQUIRED(self):
        return self._obligation == REQUIRED


class OutPort(Port):
    '''Defines the specific behavior for connecting to outports i.e. how to 
    check for downstream data types for scaling downstream ports on the canvas
    to highlight potentially valid connections.
    '''

    PortType = OutPortTYPE

    def __init__(self, nodeWidget, CanvasBackend, title, portNum, intype=None, dtype=None, ndim=None, obligation=None, menuWidget=None):
        super(OutPort, self).__init__(nodeWidget, CanvasBackend,
                                      title, portNum, intype, dtype, ndim, menuWidget)

        self.data_changed = True  # don't start in warning state

    def setDataCalled(self, called=True):
        self.data_changed = called

    def dataHasChanged(self):
        '''Used for setting port color and signaling downstream events.
        '''
        return self.data_changed

    def dataIsNone(self):
        return (self.data is None)

    def findMatchingInPorts(self, ports=None):
        if ports == None: ports = self.findOppositePorts()
        matching_ports = []

        if len(ports):
            # check data if available
            if not self.dataIsNone():
                for port in ports:
                    # skip ports on the same node
                    if port.node == self.node:
                        continue
                    if not len(port.edges()):  # dont consider ports with connections
                        if port.checkDataType(self.data):
                            matching_ports.append(port)

            # if not, just check against port spec
            else:
                for port in ports:
                    # skip ports on the same node
                    if port.node == self.node:
                        continue
                    if not len(port.edges()):  # dont consider ports with connections
                        # call checker from inport since they are
                        # the limiting factor
                        if port.checkTypeAgainstPort(self):
                            matching_ports.append(port)

        return matching_ports

    def type(self):
        return OutPort.Type

    def porttype(self):
        return OutPort.PortType

    # return a read-only ptr
    @property
    def data(self):
        return self._data

    def freeDataQuietly(self):
        '''Quietly free (remove ref) the data causing no new events.
        All connected downstream nodes must run before this is freed.
        '''
        self._data = None

    def getDataString(self):
        # PORTAUTH
        return self._GPIType.edgeTip(self._data)

    def setWidgetData(self):  # widget-ports
        val = self.menuWidget.get_val()
        # This comparison could be an expensive operation
        # not sure if ids are checked first.
        if self._data != val:
            if self.setData(val):  # check val against port type
                return True
        return False  # data didn't change

    def setData(self, out):
        # PORTAUTH
        # enforce port type
        if self.checkDataType(out):
            self._data = self._GPIType.setDataAttr(out)
            self.setDataCalled()
            return True  # SUCCESS

        elif out is None:
            #log.warn("setData(): OutPort unchanged.")
            self._data = None
            #self.setDataCalled()
            #return True  # SUCCESS

        else:
            log.warn("setData(\'"+stw(self.portTitle)+"\',...): OutPort type doesn't match data, unchanged.")

        self.setDataCalled(False)
        return False

    def setDownstreamEvents(self):
        '''Look at downstream Nodes and flag pending_event regardless of obligation.
            -Finds inports connected to THIS outport.
        '''
        for nodeObj in self.getDownstreamNodes():
            # nodeObj[node,portTitle]
            # if self.isWidgetPort():
            #    print "setDownstreamEvents(): Send widget-event down to: "+node.name
            # else:
            #    print "setDownstreamEvents(): Send event down to: "+node.name
            nodeObj[0].setEventStatus({GPI_PORT_EVENT: nodeObj[1]})
